<?php

declare(strict_types=1);

namespace app\models\connection;

use yii;
use yii\base\Model;

use app\models\Dellin as DellinModel;

use GuzzleHttp\Client as Guzzle;
use GuzzleHttp\Exception\ClientException as GuzzleException;

use Exception;
use phpDocumentor\Reflection\Types\Nullable;

class Dellin extends Model
{
    /**
     * Инстанция браузера
     */
    public static Guzzle $browser;

    /**
     * Сессия аккаунта
     */
    public static string $session;

    public function __construct($config = [])
    {
        parent::__construct($config);

        self::$browser = new Guzzle([
            'base_uri'  => 'https://api.dellin.ru/'
        ]);

        self::authorization();
    }

    // /**
    //  * Поиск городов
    //  *
    //  * @return array|null Найденные города
    //  */
    // public function searchCities(): ?array
    // {
    // $this->onReady(function () {
    //     // Запрос городов
    //     $request = $this->browser->post('/v2/public/kladr.json', [
    //         'json' => [
    //             'appkey' => yii::$app->params['dellin']['key'],

    //         ]
    //     ])
    // });
    // }

    /**
     * Рассчет доставки (расширенный)
     *
     * @param string $from Номер КЛАДР
     * @param string $to Номер КЛАДР
     * @param int $weight Вес (кг)
     * @param int $x Ширина (м)
     * @param int $y Высота (м)
     * @param int $z Длинна (м)
     * @param int $amount Количество
     *
     * @return string
     *
     * @todo Переработать "weight" (если посылок больше чем одна, то отправить вес самого тяжелого груза), хотя здесь же копии обрабатываются и вес у них один... Надо думать
     */
    public static function calcDeliveryAdvanced(string $from, string $to, int $weight, int $x, int $y, int $z, int $amount = 1): array
    {
        return self::handle(function () use ($from, $to, $weight, $x, $y, $z, $amount) {
            // Всё готово к работе

            // Рассчёт типа доставки
            if (
                $weight <= 10
                && (($x <= 5.4 && $y <= 3.9 && $z <= 3.9)
                    || ($x <= 3.9 && $y <= 5.4 && $z <= 3.9)
                    || ($x <= 3.9 && $y <= 3.9 && $z <= 5.4))
                && $x * $y * $z <= 0.1
            ) {
                // Доставка категории "small"

                $type = 'small';
            } else {
                // Доставка категории "auto"

                $type = 'auto';
            }

            // Вычисление самой крупной стороны, так как ДеловыеЛинии имеют ограничения на все три поля и у длинны оно больше всех
            if ($x > $z && $x > $y) {
                // "X" больше всех

                // Инициализация
                $width = $z;
                $height = $y;
                $length = $x;
            } else if ($y > $x && $y > $z) {
                // "Y" больше всех

                // Инициализация
                $width = $x;
                $height = $z;
                $length = $y;
            } else if ($z > $x && $z > $y) {
                // "Z" больше всех

                // Инициализация
                $width = $x;
                $height = $y;
                $length = $z;
            } else {
                echo 'Ошибка';
            }

            // Запрос
            $request = self::$browser->post('/v2/calculator.json', [
                'json' => [
                    'appkey' => yii::$app->params['dellin']['key'],
                    'sessionID' => self::$session,
                    'delivery' => [
                        'deliveryType' => [
                            'type' => $type
                        ],
                        'derival' => [
                            'produceDate' => date('Y-m-d', time() + ($settings['delivery_handle_time'] ?? 86400)),
                            'variant' => 'terminal',
                            'terminalID' => $from,

                        ],
                        'arrival' => [
                            'variant' => 'terminal',
                            'terminalID' => $to
                        ]
                    ],
                    'members' => [
                        'requester' => [
                            'role' => 'sender'
                        ]
                    ],
                    'cargo' => [
                        'quantity' => $amount,
                        'width' => $width,
                        'height' => $height,
                        'length' => $length,
                        'weight' => $weight,
                        'totalVolume' => $x * $y * $z,
                        'totalWeight' => $weight,
                        'oversizedWeight' => $weight,
                        'oversizedVolume' => $x * $y * $z
                    ]
                ]
            ]);

            if ($request->getStatusCode() === 200) {
                // Запрос прошел успешно

                // Инициализация
                $response = json_decode((string) $request->getBody(), true);

                if ($response['metadata']['status'] === 200) {
                    // Со стороны ДеловыеЛинии ошибок нет

                    return $response['data'];
                }

                throw new Exception('На стороне сервера ДеловыеЛинии какие-то проблемы, либо отправлен неверный запрос', 500);
            }

            throw new Exception('Не удалось запросить рассчёт доставки у ДеловыеЛинии', 500);
        });
    }


    /**
     * Рассчет доставки
     *
     * @param string $from Номер КЛАДР
     * @param string $to Номер КЛАДР
     *
     * @return string
     */
    public static function calcDelivery(string $from, string $to): array
    {
        return self::handle(function () use ($from, $to) {
            // Всё готово к работе

            // Запрос
            $request = self::$browser->post('/v1/micro_calc.json', [
                'json' => [
                    'appkey' => yii::$app->params['dellin']['key'],
                    'sessionID' => self::$session,
                    'derival' => [
                        'city' => $from
                    ],
                    'arrival' => [
                        'city' => $to
                    ]
                ]
            ]);


            if ($request->getStatusCode() === 200) {
                // Запрос прошел успешно

                // Инициализация
                $response = json_decode((string) $request->getBody(), true);

                if ($response['metadata']['status'] === 200) {
                    // Со стороны ДеловыеЛинии ошибок нет

                    return $response['data'];
                }

                throw new Exception('На стороне сервера ДеловыеЛинии какие-то проблемы, либо отправлен неверный запрос', 500);
            }

            throw new Exception('Не удалось запросить рассчёт доставки у ДеловыеЛинии', 500);
        });
    }

    /**
     * Импорт городов
     *
     * @return array|null Сохранённые города
     */
    // public static function importCities(): ?int
    // {
    //     return self::handle(function () {
    //         // Всё готово к работе

    //         // Запрос ссылки на файл с городами, возвращает ['hash' => string, 'url' => string]
    //         $request = self::$browser->post('/v1/public/cities.json', [
    //             'json' => [
    //                 'appkey' => yii::$app->params['dellin']['key'],
    //             ]
    //         ]);

    //         if ($request->getStatusCode() === 200) {
    //             // Запрос прошел успешно

    //             // Инициализация
    //             $response = json_decode((string) $request->getBody(), true);
    //             $hash_target = $response['hash'];
    //             $dir = YII_PATH_PUBLIC . '/../assets/import/' . date('Y-m-d', time()) . '/dellin/cities/' . (yii::$app->user->identity->_key ?? 'system') . '/';

    //             if (!file_exists($dir)) {
    //                 // Директории не существует

    //                 mkdir($dir, 0775, true);
    //             }

    //             $request = self::$browser->get($response['url'], [
    //                 'sink' => $file = $dir . time() . '.csv'
    //             ]);

    //             // Проверка хеша (оказалось это хеш запроса, бесполезный)
    //             // if ($hash_target === $hash_received = md5_file($file)) {
    //             // Удалось пройти проверку на хеши файлов

    //             // Инициализация (чтение файла)
    //             $file = fopen($file, "r");
    //             $first_raw_block = true;

    //             while ($row = fgets($file, 4096)) {
    //                 // Перебор строк

    //                 if ($first_raw_block) {
    //                     // Сработала защита от чтения первой строки файла (указываются названия колонок)

    //                     // Отключение
    //                     $first_raw_block = false;

    //                     // Пропуск цикла
    //                     continue;
    //                 }

    //                 // Инициализация
    //                 $data = explode(',', $row, 4);
    //                 $amount = 0;

    //                 // Очистка
    //                 array_walk($data, fn (&$value) => $value = trim($value, '"'));

    //                 if ($city = City::searchByDellinId($data[0])) {
    //                     // Удалось найти город в базе данных

    //                     $after_import_log = function () use ($city): void {
    //                         // Запись в журнал
    //                         $city->journal('update');

    //                         if (yii::$app->getRequest()->isConsoleRequest) {
    //                             // Вызов из терминала

    //                             echo 'Удалось перезаписать город: ' . $city->name . PHP_EOL;
    //                         }
    //                     };
    //                 } else {
    //                     // Не удалось найти город в базе данных

    //                     $city = new City();

    //                     $after_import_log = function () use ($city): void {
    //                         if (yii::$app->getRequest()->isConsoleRequest) {
    //                             // Вызов из терминала

    //                             echo 'Удалось записать город: ' . $city->name . PHP_EOL;
    //                         }
    //                     };
    //                 }

    //                 // Запись
    //                 $city->indx = $data[0];
    //                 $city->name = $data[1];
    //                 $city->code = $data[2];
    //                 $city->term = (bool) $data[3];

    //                 // Отправка в базу данных
    //                 if ($city->save()) {
    //                     // Удалось сохранить в базе данных

    //                     // Запись в журнал
    //                     $after_import_log();

    //                     // Постинкрементация счётчика
    //                     $amount++;

    //                     continue;
    //                 } else {
    //                     // Не удалось сохранить в базе данных

    //                     throw new Exception('Не удалось сохранить город ' . $data[1] . ' в базу данных', 500);
    //                 }
    //             }

    //             // Деинициализация
    //             fclose($file);

    //             return $amount;
    //             // } else {
    //             //     // Не удалось пройти проверку на соответствие хешей файлов

    //             //     throw new Exception('Хеши файлов не совпадают. Должен быть: "' . $hash_target . '", получен: "' . $hash_received . '"', 500);
    //             // }
    //         }

    //         throw new Exception('Не удалось синхронизировать данные городов с ДеловыеЛинии', 500);
    //     });
    // }


    /**
     * Импорт терминалов
     *
     * @return array|null Сохранённые терминалы
     */
    public static function importTerminals(): ?int
    {
        return self::handle(function () {
            // Всё готово к работе

            // Запрос ссылки на файл с городами, возвращает ['hash' => string, 'url' => string]
            $request = self::$browser->post('/v3/public/terminals.json', [
                'json' => [
                    'appkey' => yii::$app->params['dellin']['key'],
                ]
            ]);

            if ($request->getStatusCode() === 200) {
                // Запрос прошел успешно

                // Инициализация
                $response = json_decode((string) $request->getBody(), true);
                $dir = YII_PATH_PUBLIC . '/../assets/import/' . date('Y-m-d', time()) . '/dellin/terminals/' . (yii::$app->user->identity->_key ?? 'system') . '/';
                $amount = 0;

                if (!file_exists($dir)) {
                    // Директории не существует

                    mkdir($dir, 0775, true);
                }

                $request = self::$browser->get($response['url'], [
                    'sink' => $file = $dir . time() . '.json'
                ]);

                // Инициализация
                $terminals = json_decode(fread(fopen($file, "r"), filesize($file)), true);

                foreach ($terminals['city'] as $terminal) {
                    // Перебор городов

                    if ($model = DellinModel::searchByCityId($terminal['id'])) {
                        // Удалось найти город в базе данных

                        $after_import_log = function () use ($model): void {
                            // Запись в журнал
                            $model->journal('update');

                            if (yii::$app->getRequest()->isConsoleRequest) {
                                // Вызов из терминала

                                echo 'Удалось перезаписать терминалы города: ' . ($model->data['name'] ?? 'Неизвестно') . PHP_EOL;
                            }
                        };
                    } else {
                        // Не удалось найти город в базе данных

                        $model = new DellinModel();

                        $after_import_log = function () use ($model): void {
                            if (yii::$app->getRequest()->isConsoleRequest) {
                                // Вызов из терминала

                                echo 'Удалось записать терминалы города: ' . ($model->data['name'] ?? 'Неизвестно') . PHP_EOL;
                            }
                        };
                    }

                    // Запись
                    $model->data = $terminal;

                    // Отправка в базу данных
                    if ($model->save()) {
                        // Удалось сохранить в базе данных

                        // Запись в журнал
                        $after_import_log();

                        // Постинкрементация счётчика
                        $amount++;

                        continue;
                    } else {
                        // Не удалось сохранить в базе данных

                        throw new Exception('Не удалось сохранить терминалы города "' . ($model->data['name'] ?? 'Неизвестно') . '" в базу данных', 500);
                    }
                }

                return $amount;
            }

            throw new Exception('Не удалось синхронизировать данные городов с ДеловыеЛинии', 500);
        });
    }

    /**
     * Аутентификация и авторизация
     */
    protected static function authorization(): bool
    {
        // Аутентификация и авторизация
        $request = self::browser()->post('/v3/auth/login.json', [
            'json' => [
                'appkey' => yii::$app->params['dellin']['key'],
                'login' => yii::$app->params['dellin']['nickname'],
                'password' => yii::$app->params['dellin']['password']
            ]
        ]);

        if ($request->getStatusCode() === 200) {
            // Запрос прошел успешно

            // Инициализация
            $response = json_decode((string) $request->getBody(), true);

            if ($response['metadata']['status'] === 200) {
                // Аутентификация и авторизация пройдены успешно

                // Запись сессии
                self::$session = $response['data']['sessionID'];

                return true;
            }

            throw new Exception('Не удалось авторизироваться в ДеловыеЛинии', $response['metadata']['status']);
        }

        throw new Exception('Не удалось авторизироваться в ДеловыеЛинии', $request->getStatusCode);
    }

    /**
     * Инициализация и выполнение
     *
     * @param callable|null $function Код к выполнению
     * @param [mixed] ...$vars Параметры к нему
     *
     * @return mixed Возврат из функции
     */
    protected static function handle(callable $function = null, mixed ...$vars): mixed
    {
        try {
            if (self::browser() instanceof Guzzle) {
                // Браузер инициализирован

                if (self::authorization() && isset(self::$session)) {
                    // Аутентифицирован и авторизован

                    return $function(...$vars);
                } else {
                    throw new Exception('Аккаунт не авторизирован', 401);
                }
            } else {
                throw new Exception('Браузер не инициализирован', 500);
            }
        } catch (GuzzleException $e) {
            throw new Exception($e->getResponse()->getBody()->getContents() ?? 'Не удалось инициализировать инстанцию для работы с API ДеловыеЛинии', 500, $e->getPrevious());
        } catch (Exception $e) {
            throw new Exception($e->getMessage() ?? 'Не удалось инициализировать инстанцию для работы с API ДеловыеЛинии', 500, $e->getPrevious());
        }
    }

    /**
     * Чтение или инициализация браузера
     *
     * @return Guzzle Инстанция
     */
    protected static function browser(): Guzzle
    {
        return self::$browser ?? self::$browser = new Guzzle([
            'base_uri'  => 'https://api.dellin.ru/'
        ]);
    }
}
